var qs = require('qs');


function expressRest(opts) {
  opts                 || (opts                 = {});
  opts.maxResultsLimit || (opts.maxResultsLimit = Infinity);
  opts.resourceId      || (opts.resourceId      = '_id');
  opts.useEnvelope     || (opts.useEnvelope     = false);
  opts.propertyCase    || (opts.propertyCase    = 'camelCase');

  var resourceId = opts.resourceId;

  var getNextUrl = (req, opts) => {
    if (opts.skip + opts.limit >= opts.count) { return ''; }
    var nextPageQuery = Object.assign({}, req.query);
    nextPageQuery.$skip  = opts.skip + opts.limit;
    nextPageQuery.$limit = Math.min(opts.limit, opts.count - opts.skip);
    req.envelope     && (nextPageQuery.$envelope     = req.envelope);
    req.propertyCase && (nextPageQuery.$propertyCase = req.propertyCase);
    req.sort         && (nextPageQuery.$sort         = req.sort);
    return opts.baseUrl + '?' + qs.stringify(nextPageQuery);
  };

  var getPreviousUrl = (req, opts) => {
    if (opts.skip <= 0) { return ''; }
    var lastPageQuery = Object.assign({}, req.query);
    lastPageQuery.$skip  = Math.max(opts.skip - opts.limit, 0);
    lastPageQuery.$limit = Math.min(opts.skip + opts.limit, opts.count) - opts.skip;
    req.envelope     && (lastPageQuery.$envelope     = req.envelope);
    req.propertyCase && (lastPageQuery.$propertyCase = req.propertyCase);
    req.sort         && (lastPageQuery.$sort         = req.sort);
    return opts.baseUrl + '?' + qs.stringify(lastPageQuery);
  };

  var calcOpts = (req, opts) => {
    opts             || (opts             = {});
    opts.sort        || (opts.sort        = req.sort);
    opts.skip        || (opts.skip        = req.skip);
    opts.limit       || (opts.limit       = req.limit);
    opts.count       || (opts.count       = 0);
    opts.baseUrl     || (opts.baseUrl     = req.originalUrl.split('?')[0]);
    opts.useEnvelope || (opts.useEnvelope = req.envelope);
    opts.nextUrl     || (opts.nextUrl     = getNextUrl(req, opts));
    opts.lastUrl     || (opts.lastUrl     = getPreviousUrl(req, opts));
    opts.location    || (opts.location    = opts.baseUrl);
    return opts;
  };

  var toDelimitedCase = (str, propertyCase) => {
    switch (propertyCase) {

      case 'snake_case':
        return str
          .replace(/([\w\d])-([\w\d])/g, (_, m1, m2) => m1 + '_' + m2)
          .replace(/([\w\d])?([A-Z])/g, (_, m1, m2) => m1 ? m1 + '_' + m2.toLowerCase() : m2.toLowerCase());

      case 'kebab-case':
        return str
          .replace(/([\w\d])_([\w\d])/g, (_, m1, m2) => m1 + '-' + m2)
          .replace(/([\w\d])?([A-Z])/g, (_, m1, m2) => m1 ? m1 + '-' + m2.toLowerCase() : m2.toLowerCase());

      case 'camelCase':
      default:
        return str.replace(/([\w\d])[-_]([\w\d])/g, (_, m1, m2) => m1 + m2.toUpperCase());
    }
  };

  var setPropertyCase = (ctx, propertyCase) => {
    for (var currentPropName in ctx) {
      if (!ctx.hasOwnProperty(currentPropName)) { continue; }

      if (
        ctx[currentPropName] &&
        typeof ctx[currentPropName] === 'object' && (
          ctx[currentPropName].constructor === Object ||
          ctx[currentPropName].constructor === Array
      )) {
        setPropertyCase(ctx[currentPropName], propertyCase);
      }

      var newPropName = toDelimitedCase(currentPropName, propertyCase);
      var val = ctx[currentPropName];
      delete ctx[currentPropName];
      ctx[newPropName] = val;
    }
  };

  var sendCreateOrAccepted = function(req, res, type, item, opts) {
    if (item.toJSON) { item = item.toJSON(); }

    opts           = calcOpts(req, opts);
    opts.location += '/' + item[resourceId];

    res.status(type === 'created' ? 201 : 202);

    if (opts.useEnvelope) {
      return res.json({
        item    : item,
        location: opts.location
      });
    }

    res.set('location', opts.location);

    return res.json(item);
  };

  var middleware = (req, res, next) => {

    if (req.query.$envelope !== undefined) {
      req.envelope = req.query.$envelope;
    } else if (req.get('x-envelope') !== undefined) {
      req.envelope = req.get('x-envelope');
    } else {
      req.envelope = opts.useEnvelope;
    }
    req.envelope = req.envelope && req.envelope !== 'false' && req.envelope !== '0';

    if (req.query.$propertyCase !== undefined) {
      req.propertyCase = req.query.$propertyCase;
    } else if (req.get('x-property-case') !== undefined) {
      req.propertyCase = req.get('x-property-case');
    } else {
      req.propertyCase = opts.propertyCase;
    }

    req.sort  = req.query.$sort;
    req.skip  = req.query.$skip  ? parseFloat(req.query.$skip) : 0;
    req.limit = req.query.$limit ?
      Math.min(parseFloat(req.query.$limit), opts.maxResultsLimit) :
      opts.maxResultsLimit;
    delete req.query.$envelope;
    delete req.query.$propertyCase;
    delete req.query.$sort;
    delete req.query.$skip;
    delete req.query.$limit;

    setPropertyCase(req.query, 'camelCase');
    req.body && setPropertyCase(req.body, 'camelCase');

    var originalJson = res.json;
    res.json = function(obj) {
      setPropertyCase(obj, req.propertyCase);
      return originalJson.call(this, obj);
    };

    res.sendFound = function(item, opts) {
      if (item.toJSON) { item = item.toJSON(); }

      opts = calcOpts(req, opts);

      if (opts.useEnvelope) {
        return this.json({
          item    : item,
          location: opts.location
        });
      }

      this.set('location', opts.location);

      return this.json(item);
    };

    res.sendQueried = function(items, count, opts) {
      if (items.toJSON) { items = items.toJSON(); }

      opts || (opts = {});
      opts.count = count;
      opts = calcOpts(req, opts);

      if (opts.useEnvelope) {
        var envelope = { items: items, count: count };
        opts.nextUrl && (envelope.nextUrl = opts.nextUrl);
        opts.lastUrl && (envelope.lastUrl = opts.lastUrl);
        return this.json(envelope);
      }

      var links = {};
      opts.nextUrl && (links.next = opts.nextUrl);
      opts.lastUrl && (links.last = opts.lastUrl);
      this.links(links);

      this.set('x-total-count', count);

      return this.json(items);
    };

    res.sendCreated = function(item, opts) {
      return sendCreateOrAccepted(req, res, 'created', item, opts);
    };

    res.sendAccepted = function(item, opts) {
      return sendCreateOrAccepted(req, res, 'accepted', item, opts);
    };

    next(null);
  };

  return middleware;
}

module.exports = expressRest;
